def add_user(login, password, name, roleId, dev):
    roleId = int(roleId)
    print(f"add_user {login=} {password=} {name=} {roleId=} {dev=}")
    from bafser import db_session, Role
    from bafser.data.user import get_user_table

    db_session.global_init(dev)
    db_sess = db_session.create_session()
    User = get_user_table()
    user_admin = User.get_admin(db_sess)
    existing = User.get_by_login(db_sess, login, includeDeleted=True)
    if existing:
        print(f"User with login [{login}] already exist")
        return
    role = db_sess.get(Role, roleId)
    if not role:
        print(f"Role with id [{roleId}] does not exist")
        return

    User.new(user_admin, login, password, name, [roleId])

    print("User added")


def run(args):
    if not (len(args) == 4 or (len(args) == 5 and args[-1] == "dev")):
        print("add_user: login password name roleId [dev]")
    else:
        add_user(args[0], args[1], args[2], args[3], args[-1] == "dev")



from datetime import datetime


def parse_date(date: str):
    try:
        if date[-1] == "Z":
            date = date[:-1]
        return datetime.fromisoformat(date), True
    except Exception:
        return None, False



import os
from flask import current_app, make_response, send_file


def create_file_response(fpath: str, ftype: str, fname: str):
    if not os.path.exists(fpath):
        return make_response("", 404)
    max_age = current_app.config["CACHE_MAX_AGE"]
    response = send_file(fpath)
    response.headers.set("Content-Type", ftype)
    response.headers.set("Content-Disposition", "inline", filename=fname)
    response.headers.set("Cache-Control", f"public,max-age={max_age},immutable")
    return response



from datetime import datetime, timedelta, timezone
from typing import Callable, Literal, Union
import json
import logging
import os
import time
import traceback

from flask import Flask, Response, abort, g, make_response, redirect, request, send_from_directory
from flask_jwt_extended import JWTManager, create_access_token, get_jwt, get_jwt_identity, set_access_cookies, verify_jwt_in_request
from urllib.parse import quote
from sqlalchemy.orm import Session


from . import db_session
from .alembic import alembic_upgrade
from .logger import get_logger_requests, setLogging
from .utils import get_json, get_secret_key, get_secret_key_rnd, randstr, register_blueprints, response_msg
import bafser_config


class AppConfig():
    data_folders: list[tuple[str, str]] = []
    config: list[tuple[str, str]] = []

    def __init__(self,
                 FRONTEND_FOLDER="build",
                 IMAGES_FOLDER="images",
                 JWT_ACCESS_TOKEN_EXPIRES: Union[Literal[False], timedelta] = False,
                 JWT_ACCESS_TOKEN_REFRESH: Union[Literal[False], timedelta] = timedelta(minutes=30),
                 CACHE_MAX_AGE=31536000,
                 MESSAGE_TO_FRONTEND="",
                 STATIC_FOLDERS: list[str] = ["/static/", "/fonts/", "/_next/"],
                 DEV_MODE=False,
                 DELAY_MODE=False,
                 PAGE404="index.html",
                 ):
        self.FRONTEND_FOLDER = FRONTEND_FOLDER
        self.IMAGES_FOLDER = IMAGES_FOLDER
        self.JWT_ACCESS_TOKEN_EXPIRES = JWT_ACCESS_TOKEN_EXPIRES
        self.JWT_ACCESS_TOKEN_REFRESH = JWT_ACCESS_TOKEN_REFRESH
        self.CACHE_MAX_AGE = CACHE_MAX_AGE
        self.MESSAGE_TO_FRONTEND = MESSAGE_TO_FRONTEND
        self.STATIC_FOLDERS = STATIC_FOLDERS
        self.DEV_MODE = DEV_MODE
        self.DELAY_MODE = DELAY_MODE
        self.PAGE404 = PAGE404
        self.add_data_folder("IMAGES_FOLDER", IMAGES_FOLDER)
        self.add("CACHE_MAX_AGE", CACHE_MAX_AGE)

    def add(self, key: str, value: str):
        self.config.append((key, value))
        return self

    def add_data_folder(self, key: str, path: str):
        self.add(key, path)
        self.data_folders.append((key, path))
        return self

    def add_secret_key(self, key: str, path: str):
        self.add(key, get_secret_key(path))
        return self

    def add_secret_key_env(self, key: str, envname: str = None, default: str = None):
        if envname is None:
            envname = key
        v = os.environ.get(envname, default)
        if v is None:
            raise Exception(f"env var not set: {envname}")
        self.add(key, v)
        return self

    def add_secret_key_rnd(self, key: str, path: str):
        self.add(key, get_secret_key_rnd(path))
        return self


def create_app(import_name: str, config: AppConfig):
    setLogging()
    logreq = get_logger_requests()
    app = Flask(import_name, static_folder=None)
    app.config["JWT_TOKEN_LOCATION"] = ["cookies"]
    app.config["JWT_SECRET_KEY"] = get_secret_key_rnd(bafser_config.jwt_key_file_path)
    app.config["JWT_ACCESS_TOKEN_EXPIRES"] = config.JWT_ACCESS_TOKEN_EXPIRES
    app.config["JWT_COOKIE_CSRF_PROTECT"] = False
    app.config["JWT_SESSION_COOKIE"] = False
    app.secret_key = app.config["JWT_SECRET_KEY"]
    for (key, path) in config.config:
        app.config[key] = path

    jwt_manager = JWTManager(app)

    register_blueprints(app)

    for (_, path) in config.data_folders:
        if not os.path.exists(path):
            os.makedirs(path)

    def run(
        run_server: bool,
        init_db: Callable[[Session, AppConfig], None] = None,
        init_dev_values: Callable[[Session, AppConfig], None] = None,
        port=5000,
        host="127.0.0.1"
    ):

        if bafser_config.use_alembic:
            alembic_upgrade(config.DEV_MODE)
        init_database(init_db, init_dev_values)

        if run_server:
            print(f"Starting on http://{host}:{port}")
            if config.DELAY_MODE:
                print("Delay for requests is enabled")
            app.run(debug=config.DEV_MODE, port=port, host=host)

    def init_database(init_db: Callable[[Session, AppConfig], None], init_dev_values: Callable[[Session, AppConfig], None]):
        from . import Role, UserBase
        from .data.db_state import DBState
        db_session.global_init(config.DEV_MODE)
        with db_session.create_session() as db_sess:
            is_initialized = DBState.is_initialized(db_sess)
            if not is_initialized:
                print("initialize database")
                Role.update_roles_permissions(db_sess)
                UserBase._create_admin(db_sess)
                if init_db is not None:
                    print("init_db")
                    init_db(db_sess, config)
                if config.DEV_MODE and init_dev_values is not None:
                    print("init_dev_values")
                    init_dev_values(db_sess, config)
                DBState.mark_as_initialized(db_sess)

            if not config.DEV_MODE:
                change_admin_default_pwd(db_sess)

    def change_admin_default_pwd(db_sess: Session):
        from . import UserBase
        admin = UserBase.get_by_login(db_sess, "admin", includeDeleted=True)
        if admin is not None and admin.check_password("admin"):
            admin.set_password(randstr(16))
            db_sess.commit()
        db_sess.close()

    @app.before_request
    def before_request():
        g.json = get_json(request)
        g.req_id = randstr(4)
        try:
            verify_jwt_in_request()
            jwt_identity = get_jwt_identity()
        except Exception:
            jwt_identity = None
        if jwt_identity and isinstance(jwt_identity, (list, tuple)) and len(jwt_identity) == 2:
            g.userId = jwt_identity[0]
        if request.path.startswith(bafser_config.api_url):
            try:
                if g.json[1]:
                    if "password" in g.json[0]:
                        password = g.json[0]["password"]
                        g.json[0]["password"] = "***"
                        data = json.dumps(g.json[0])[:512]
                        g.json[0]["password"] = password
                    else:
                        data = json.dumps(g.json[0])[:512]
                    logreq.info("Request;;%(data)s", {"data": data})
                else:
                    logreq.info("Request")
            except Exception as x:
                logging.error("Request logging error: %s", x)

        if config.DELAY_MODE:
            time.sleep(0.5)

    @app.after_request
    def after_request(response: Response):
        if request.path.startswith(bafser_config.api_url):
            try:
                if response.content_type == "application/json":
                    logreq.info("Response;%s;%s", response.status_code, str(response.data)[:512])
                else:
                    logreq.info("Response;%s", response.status_code)
            except Exception as x:
                logging.error("Request logging error: %s", x)

        response.set_cookie("MESSAGE_TO_FRONTEND", quote(config.MESSAGE_TO_FRONTEND))

        if config.JWT_ACCESS_TOKEN_REFRESH:
            try:
                exp_timestamp = get_jwt()["exp"]
                now = datetime.now(timezone.utc)
                target_timestamp = datetime.timestamp(now + config.JWT_ACCESS_TOKEN_REFRESH)
                if target_timestamp > exp_timestamp:
                    access_token = create_access_token(identity=get_jwt_identity())
                    set_access_cookies(response, access_token)
            except (RuntimeError, KeyError):
                # Case where there is not a valid JWT
                pass

        return response

    @app.route("/", defaults={"path": ""})
    @app.route("/<path:path>")
    def frontend(path):
        if request.path.startswith(bafser_config.api_url):
            abort(404)

        if path == "":
            fname = "index.html"
        elif os.path.exists(config.FRONTEND_FOLDER + "/" + path):
            fname = path
        elif os.path.exists(config.FRONTEND_FOLDER + "/" + path + ".html"):
            fname = path + ".html"
        else:
            fname = config.PAGE404

        res = send_from_directory(config.FRONTEND_FOLDER, fname)
        if any(request.path.startswith(path) for path in config.STATIC_FOLDERS):
            res.headers.set("Cache-Control", f"public,max-age={config.CACHE_MAX_AGE},immutable")
        else:
            res.headers.set("Cache-Control", "public,max-age=60,stale-while-revalidate=600,stale-if-error=14400")
        return res

    @app.errorhandler(404)
    def not_found(error):
        if request.path.startswith(bafser_config.api_url):
            return response_msg("Not found", 404)
        return make_response("Страница не найдена", 404)

    @app.errorhandler(405)
    def method_not_allowed(error):
        return response_msg("Method Not Allowed", 405)

    @app.errorhandler(415)
    def unsupported_media_type(error):
        return response_msg("Unsupported Media Type", 415)

    @app.errorhandler(403)
    def no_permission(error):
        return response_msg("No permission", 403)

    @app.errorhandler(500)
    @app.errorhandler(Exception)
    def internal_server_error(error):
        print(error)
        logging.error("%s\n%s", error, traceback.format_exc())
        if request.path.startswith(bafser_config.api_url):
            return response_msg("Internal Server Error", 500)
        return make_response("Произошла ошибка", 500)

    @app.errorhandler(401)
    def unauthorized(error):
        if request.path.startswith(bafser_config.api_url):
            return response_msg("Unauthorized", 401)
        return redirect(bafser_config.login_page_url)

    @jwt_manager.expired_token_loader
    def expired_token_loader(jwt_header, jwt_data):
        return response_msg("The JWT has expired", 401)

    @jwt_manager.invalid_token_loader
    def invalid_token_loader(error):
        return response_msg("Invalid JWT", 401)

    @jwt_manager.unauthorized_loader
    def unauthorized_loader(error):
        return response_msg("Unauthorized", 401)

    return app, run



from typing import Type, TypedDict

from ..utils.get_all_vars import get_all_fields


Roles: "Type[RolesBase]" = None


def get_roles():
    if Roles is None:
        raise Exception("[bafser] No class inherited from RolesBase")
    return Roles


class RoleDesc(TypedDict):
    name: str
    operations: list[tuple[str, str]]


TRole = int
TRoles = dict[TRole, RoleDesc]


class RolesBase:
    admin = 1

    ROLES: TRoles = {}

    def __init_subclass__(cls, **kwargs):
        global Roles
        Roles = cls
        fields = list(get_all_fields(cls()))
        same = []
        for i in range(len(fields)):
            for k in range(i + 1, len(fields)):
                if fields[i][1] == fields[k][1]:
                    same.append((fields[i][0], fields[k][0], fields[i][1]))
        if len(same) > 0:
            raise Exception(f"[bafser] Same ids for Role: {'; '.join(f'{n1} and {n2} is {v}' for (n1, n2, v) in same)}")

    @staticmethod
    def get_all():
        ROLES = get_roles().ROLES
        return [(RolesBase.admin, "Admin")] + [(id, ROLES[id]["name"]) for id in ROLES.keys()]



import os
import re
from alembic import command
from alembic.script import ScriptDirectory
from ..alembic import create_alembic_config

import bafser_config


def init(args):
    print("alembic init")

    os.makedirs(bafser_config.migrations_folder, exist_ok=True)
    os.makedirs(os.path.join(bafser_config.migrations_folder, "versions"), exist_ok=True)
    with open(os.path.join(bafser_config.migrations_folder, "env.py"), "w", encoding="utf8") as f:
        f.write("from bafser.alembic import run\n\nrun()\n")
    with open(os.path.join(bafser_config.migrations_folder, "script.py.mako"), "w", encoding="utf8") as f:
        f.write(script_py_mako)


def revision(args):
    alembic_cfg = create_alembic_config(dev=True)
    if len(args) == 0:
        script = ScriptDirectory.from_config(alembic_cfg)
        maxV = 0
        for s in os.scandir(script.versions):
            m = re.match("\\d{4}_\\d{2}_\\d{2}_[a-z\\d]+_v(\\d+)\\.py", s.name)
            if not m:
                continue
            maxV = max(maxV, int(m.group(1)))
        name = f"v{maxV + 1}"
    else:
        name = args[0]
    print(f"alembic revision --autogenerate -m \"{name}\"")
    command.revision(alembic_cfg, name, True)


def run(args):
    scripts = [
        ("init", "create folders and files", init),
        ("revision", "[name] : autogenerate migration script", revision),
    ]

    if len(args) == 0 or args[0] not in map(lambda v: v[0], scripts):
        ml = max(map(lambda v: len(v[0]), scripts))
        ml2 = max(map(lambda v: len(v[1]), scripts))
        l = ml + ml2 + 5
        t = " Scripts "
        l2 = (l - len(t))
        print("-" * (l2 // 2) + t + "-" * (l2 // 2 + l2 % 2))
        print("\n".join(map(lambda v: f"{' ' * (ml - len(v[0]))} {v[0]} : {v[1]}", scripts)))
        print("-" * l)
        return

    for s in scripts:
        if args[0] == s[0]:
            s[2](args[1:])


script_py_mako = '''"""${message}

Revision ID: ${up_revision}
Revises: ${down_revision | comma,n}
Create Date: ${create_date}

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
${imports if imports else ""}

# revision identifiers, used by Alembic.
revision: str = ${repr(up_revision)}
down_revision: Union[str, None] = ${repr(down_revision)}
branch_labels: Union[str, Sequence[str], None] = ${repr(branch_labels)}
depends_on: Union[str, Sequence[str], None] = ${repr(depends_on)}


def upgrade() -> None:
    ${upgrades if upgrades else "pass"}


def downgrade() -> None:
    ${downgrades if downgrades else "pass"}
'''



import os


def get_db_path(path: str):
    if not path.startswith("ENV:"):
        return path
    path = path[len("ENV:"):]
    v = os.environ.get(path)
    if v is None:
        raise Exception(f"env var not set: {path}")
    return v



from datetime import datetime
from typing import TYPE_CHECKING

from sqlalchemy.orm import Session
from sqlalchemy_serializer import SerializerMixin
import sqlalchemy as sa

if TYPE_CHECKING:
    from . import UserBase


class TableBase(SerializerMixin):
    __table_args__ = {"mysql_charset": "utf8mb4", "mysql_collate": "utf8mb4_unicode_ci"}

    if TYPE_CHECKING:
        __tablename__: str

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def get_dict(self):
        return {}


class IdMixin:
    id = sa.Column(sa.Integer, primary_key=True, unique=True, autoincrement=True)

    @classmethod
    def get(cls, db_sess: Session, id: int):
        return db_sess.get(cls, id)

    @classmethod
    def all(cls, db_sess: Session):
        return db_sess.query(cls).all()

    def __repr__(self):
        return f"<{self.__class__.__name__}> [{self.id}]"


class ObjMixin(IdMixin):
    deleted = sa.Column(sa.Boolean, sa.DefaultClause("0"), nullable=False)

    @classmethod
    def query(cls, db_sess: Session, includeDeleted=False):
        items = db_sess.query(cls)
        if not includeDeleted:
            items = items.filter(cls.deleted == False)
        return items

    @classmethod
    def get(cls, db_sess: Session, id: int, includeDeleted=False):
        obj = db_sess.get(cls, id)
        if obj is None or (not includeDeleted and obj.deleted):
            return None
        return obj

    @classmethod
    def all(cls, db_sess: Session, includeDeleted=False):
        return cls.query(db_sess, includeDeleted).all()

    def delete(self, actor: "UserBase", commit=True, now: datetime = None, db_sess: Session = None):
        from . import Log
        self.deleted = True
        Log.deleted(self, actor, now=now, commit=commit, db_sess=db_sess)

    def restore(self, actor: "UserBase", commit=True, now: datetime = None, db_sess: Session = None):
        from . import Log
        self.deleted = False
        Log.restored(self, actor, now=now, commit=commit, db_sess=db_sess)


class SingletonMixin:
    id = sa.Column(sa.Integer, primary_key=True, unique=True, autoincrement=True)

    @classmethod
    def get(cls, db_sess: Session):
        obj = db_sess.get(cls, 1)
        if obj:
            return obj
        obj = cls(id=1)
        obj.init()
        db_sess.add(obj)
        db_sess.commit()
        return obj

    def init(self):
        pass



from functools import wraps
from flask import abort
from flask_jwt_extended import get_jwt_identity, verify_jwt_in_request
from sqlalchemy.orm import Session


def use_user_optional():
    from ..data.user import get_user_table

    def wrapper(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            if "db_sess" not in kwargs:
                abort(500, "use_user_optional: no db_sess")
            db_sess: Session = kwargs["db_sess"]

            try:
                verify_jwt_in_request()
                jwt_identity = get_jwt_identity()
                if (not isinstance(jwt_identity, list) and not isinstance(jwt_identity, tuple)) or len(jwt_identity) != 2:
                    raise Exception()

                user = get_user_table().get(db_sess, jwt_identity[0])
                if not user:
                    raise Exception()
                if user.password != jwt_identity[1]:
                    raise Exception()
            except Exception:
                user = None

            return fn(*args, **kwargs, user=user)

        return decorator

    return wrapper



from flask import abort, g

from . import response_msg


def get_json_list_from_req() -> list:
    values, is_json = g.json
    if not is_json:
        abort(response_msg("body is not json", 415))

    if not isinstance(values, list):
        abort(response_msg("body is not json list", 400))

    return values



from flask import Blueprint

from bafser import use_db_session
from sqlalchemy.orm import Session
from test.data.user import User


blueprint = Blueprint("index", __name__)


@blueprint.get("/api/user")
@use_db_session()
def index(db_sess: Session):
    u = User.get_admin(db_sess)
    return {"name": u.login}



from functools import wraps
from typing import Callable, TypeVar
from flask import abort

TCheckPermissionFn = TypeVar("TCheckPermissionFn", bound=Callable)


def create_permission_required_decorator(check_permission: TCheckPermissionFn):
    def permission_required(*args1, **kwargs1):
        def wrapper(fn):
            @wraps(fn)
            def decorator(*args2, **kwargs2):
                if "db_sess" not in kwargs2:
                    abort(500, "permission_required: no db_sess")
                if "user" not in kwargs2:
                    abort(500, "permission_required: no user")

                kwargs1["kwargs"] = kwargs2
                if not check_permission(*args1, **kwargs1):
                    abort(403)

                return fn(*args2, **kwargs2)
            return decorator
        return wrapper
    r: TCheckPermissionFn = permission_required
    return r


@create_permission_required_decorator
def permission_required(*operations: tuple[str, str], kwargs: dict = None):
    from .. import UserBase
    # db_sess: Session = kwargs["db_sess"]
    user: UserBase = kwargs["user"]

    for operation in operations:
        if not user.check_permission(operation):
            return False

    return True


@create_permission_required_decorator
def permission_required_any(*operations: tuple[str, str], kwargs: dict = None):
    from .. import UserBase
    # db_sess: Session = kwargs["db_sess"]
    user: UserBase = kwargs["user"]

    for operation in operations:
        if user.check_permission(operation):
            return True

    return False



from typing import Type

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import FunctionFilter
from sqlalchemy.sql.functions import Function
import sqlalchemy as sa
import sqlalchemy.ext.declarative as dec
import sqlalchemy.orm as orm

from .table_base import TableBase
from .utils import import_all_tables, create_folder_for_file, get_db_path
import bafser_config


convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}
SqlAlchemyBase: Type[TableBase] = dec.declarative_base(cls=TableBase)
SqlAlchemyBase.metadata = sa.MetaData(naming_convention=convention)

__factory = None


def global_init(dev: bool):
    global __factory

    if __factory:
        return

    if dev:
        create_folder_for_file(bafser_config.db_dev_path)
        conn_str = f"sqlite:///{bafser_config.db_dev_path}?check_same_thread=False"
    else:
        db_path = get_db_path(bafser_config.db_path)
        if bafser_config.db_mysql:
            conn_str = f"mysql+pymysql://{db_path}?charset=UTF8mb4"
        else:
            create_folder_for_file(db_path)
            conn_str = f"sqlite:///{db_path}?check_same_thread=False"
    print(f"Connecting to the database at {conn_str}")

    engine = sa.create_engine(conn_str, echo=bafser_config.sql_echo, pool_pre_ping=True)
    __factory = orm.sessionmaker(bind=engine)

    import_all_tables()

    SqlAlchemyBase.metadata.create_all(engine)


def create_session() -> orm.Session:
    return __factory()


# @sa.event.listens_for(sa.engine.Engine, 'connect')
# def sqlite_engine_connect(dbapi_conn, connection_record):
#     dbapi_conn.create_function('lower', 1, str.lower)


@compiles(FunctionFilter, 'mysql')
def compile_functionfilter_mysql(element, compiler, **kwgs):
    # Support unary functions only
    arg0, = element.func.clauses

    new_func = Function(
        element.func.name,
        sa.case([(element.criterion, arg0)]),
        packagenames=element.func.packagenames,
        type_=element.func.type,
        bind=element.func._bind)

    return new_func._compiler_dispatch(compiler, **kwgs)



db_dev_path = "test/db/dev.db"
db_path = "username:UR2hqJDbSfQ@username.mysql.pythonanywhere-services.com/username$default"
# db_path = "ENV:DBPATH"
db_mysql = True
sql_echo = False

use_alembic = True
migrations_folder = "test/alembic"

log_info_path = "test/logs/log_info.csv"
log_requests_path = "test/logs/log_requests.csv"
log_errors_path = "test/logs/log_errors.log"
log_frontend_path = "test/logs/log_frontend.log"

jwt_key_file_path = "test/secret_key_jwt.txt"

login_page_url = "/login"
api_url = "/api/"

blueprints_folder = "test/blueprints"
data_tables_folder = "test/data"



from datetime import datetime
from typing import Any, Union

from sqlalchemy import Column, DateTime, orm, Integer, String, JSON
from sqlalchemy.orm import Session

from .. import SqlAlchemyBase, TableBase, UserBase, IdMixin, get_datetime_now

FieldName = str
NewValue = Any
OldValue = Any


class Log(SqlAlchemyBase, IdMixin):
    __tablename__ = "Log"

    date = Column(DateTime, nullable=False)
    actionCode = Column(String(16), nullable=False)
    userId = Column(Integer, nullable=False)
    userName = Column(String(64), nullable=False)
    tableName = Column(String(16), nullable=False)
    recordId = Column(Integer, nullable=False)
    changes = Column(JSON, nullable=False)

    def __repr__(self):
        return f"<Log> [{self.id}] {self.date} {self.actionCode}"

    def get_dict(self):
        return self.to_dict(only=("id", "date", "actionCode", "userId", "userName", "tableName", "recordId", "changes"))

    @staticmethod
    def added(
        record: TableBase,
        actor: Union[UserBase, None],
        changes: list[tuple[FieldName, NewValue]],
        now: datetime = None,
        commit=True,
        db_sess: Session = None,
    ):
        if actor is None:
            actor = UserBase.get_fake_system()
        db_sess = db_sess if db_sess else Session.object_session(actor)
        if now is None:
            now = get_datetime_now()
        log = Log(
            date=now,
            actionCode=Actions.added,
            userId=actor.id,
            userName=actor.name,
            tableName=record.__tablename__,
            recordId=-1,
            changes=list(map(lambda v: (v[0], None, v[1]), changes))
        )
        db_sess.add(log)
        if isinstance(record, IdMixin):
            if record.id is not None:
                log.recordId = record.id
            elif commit:
                db_sess.commit()
                log.recordId = record.id
        if commit:
            db_sess.commit()
        return log

    @staticmethod
    def updated(
        record: TableBase,
        actor: Union[UserBase, None],
        changes: list[tuple[FieldName, OldValue, NewValue]],
        now: datetime = None,
        commit=True,
        db_sess: Session = None,
    ):
        if actor is None:
            actor = UserBase.get_fake_system()
        db_sess = db_sess if db_sess else Session.object_session(actor)
        if now is None:
            now = get_datetime_now()
        log = Log(
            date=now,
            actionCode=Actions.updated,
            userId=actor.id,
            userName=actor.name,
            tableName=record.__tablename__,
            recordId=record.id if isinstance(record, IdMixin) else -1,
            changes=changes
        )
        db_sess.add(log)
        if commit:
            db_sess.commit()
        return log

    @staticmethod
    def deleted(
        record: TableBase,
        actor: Union[UserBase, None],
        changes: list[tuple[FieldName, OldValue]] = [],
        now: datetime = None,
        commit=True,
        db_sess: Session = None,
    ):
        if actor is None:
            actor = UserBase.get_fake_system()
        db_sess = db_sess if db_sess else Session.object_session(actor)
        if now is None:
            now = get_datetime_now()
        log = Log(
            date=now,
            actionCode=Actions.deleted,
            userId=actor.id,
            userName=actor.name,
            tableName=record.__tablename__,
            recordId=record.id if isinstance(record, IdMixin) else -1,
            changes=list(map(lambda v: (v[0], v[1], None), changes))
        )
        db_sess.add(log)
        if commit:
            db_sess.commit()
        return log

    @staticmethod
    def restored(
        record: TableBase,
        actor: Union[UserBase, None],
        changes: list[tuple[FieldName, OldValue, NewValue]] = [],
        now: datetime = None,
        commit=True,
        db_sess: Session = None,
    ):
        if actor is None:
            actor = UserBase.get_fake_system()
        db_sess = db_sess if db_sess else Session.object_session(actor)
        if now is None:
            now = get_datetime_now()
        log = Log(
            date=now,
            actionCode=Actions.restored,
            userId=actor.id,
            userName=actor.name,
            tableName=record.__tablename__,
            recordId=record.id if isinstance(record, IdMixin) else -1,
            changes=changes
        )
        db_sess.add(log)
        if commit:
            db_sess.commit()
        return log


class Actions:
    added = "added"
    updated = "updated"
    deleted = "deleted"
    restored = "restored"



from typing import Type

from sqlalchemy import Column, String

from .. import SqlAlchemyBase
from ..utils import get_all_values


class Operation(SqlAlchemyBase):
    __tablename__ = "Operation"

    id = Column(String(32), primary_key=True, unique=True)
    name = Column(String(32), nullable=False)

    def __repr__(self):
        return f"<Operation> [{self.id}] {self.name}"

    def get_dict(self):
        return self.to_dict(only=("id", "name"))


class OperationsBase:
    @classmethod
    def get_all(cls):
        return list(get_all_values(cls()))

    def __init_subclass__(cls, **kwargs):
        global Operations
        Operations = cls


Operations: Type[OperationsBase] = None


def get_operations():
    if Operations is None:
        raise Exception("[bafser] No class inherited from OperationsBase")
    return Operations



from flask import Request

is_json = bool


def get_json(request: Request) -> tuple[object, is_json]:
    if not request.is_json:
        return None, False
    try:
        json = request.get_json()
        return json, True
    except Exception:
        return None, False



from datetime import datetime
from typing import Any, TypedDict, Union
import base64
import os

from flask import current_app
from sqlalchemy import Column, String, orm, ForeignKey, Integer, DateTime
from sqlalchemy.orm import Session

from .. import SqlAlchemyBase, ObjMixin, UserBase, Log, get_json_values, get_datetime_now, create_file_response
from ._tables import TablesBase


class ImageJson(TypedDict):
    data: str
    name: str


TError = str
TFieldName = str
TValue = Any


class Image(SqlAlchemyBase, ObjMixin):
    __tablename__ = TablesBase.Image

    name = Column(String(128), nullable=False)
    type = Column(String(16), nullable=False)
    creationDate = Column(DateTime, nullable=False)
    deletionDate = Column(DateTime, nullable=True)
    createdById = Column(Integer, ForeignKey("User.id"), nullable=False)

    creator = orm.relationship("User")

    @classmethod
    def new(cls, creator: UserBase, json: ImageJson) -> Union[tuple[None, TError], tuple["Image", None]]:
        (data, name), values_error = get_json_values(json, "data", "name")
        if values_error:
            return None, values_error

        data_splited = data.split(',')
        if len(data_splited) != 2:
            return None, "img data is not base64"

        img_header, img_data = data_splited
        img_header_splited = img_header.split(";")
        if len(img_header_splited) != 2 or img_header_splited[1] != "base64":
            return None, "img data is not base64"

        img_header_splited_splited = img_header_splited[0].split(":")
        if len(img_header_splited_splited) != 2:
            return None, "img data is not base64"
        mimetype = img_header_splited_splited[1]

        if mimetype not in ["image/png", "image/jpeg", "image/gif"]:
            return None, "img mimetype is not in [image/png, image/jpeg, image/gif]"

        type = mimetype.split("/")[1]

        db_sess = Session.object_session(creator)
        now = get_datetime_now()
        img, add_changes, err = cls._new(creator, json, {"name": name, "type": type, "createdById": creator.id, "creationDate": now})
        if err:
            return None, err
        db_sess.add(img)
        db_sess.commit()

        path = img.get_path()
        with open(path, "wb") as f:
            f.write(base64.b64decode(img_data + '=='))

        Log.added(img, creator, [
            ("name", img.name),
            ("type", img.type),
            ("creationDate", img.creationDate.isoformat()),
            ("createdById", img.createdById),
            *add_changes,
        ], now)

        return img, None

    @staticmethod
    def _new(creator: UserBase, json: ImageJson, image_kwargs: dict) -> \
            Union[tuple[None, None, TError], tuple["Image", list[tuple[TFieldName, TValue]], None]]:
        img = Image(**image_kwargs)
        return img, [], None

    def create_file_response(self):
        return create_file_response(self.get_path(), f"image/{self.type}", self.get_filename())

    def delete(self, actor: UserBase, commit=True, now: datetime = None, db_sess: Session = None):
        now = get_datetime_now() if now is None else now
        self.deletionDate = now
        super().delete(actor, commit, now, db_sess)

    def restore(self, actor: UserBase, commit=True, now: datetime = None, db_sess: Session = None):
        if not os.path.exists(self.get_path()):
            return False
        super().restore(actor, commit, now, db_sess)
        return True

    def get_path(self):
        return os.path.join(current_app.config["IMAGES_FOLDER"], f"{self.id}.{self.type}")

    def get_filename(self):
        return self.name + "." + self.type

    def get_dict(self):
        return self.to_dict(only=("name", "type", "creationDate", "deletionDate", "createdById"))



from typing import Any, Union
from flask import abort, g

from . import get_json_values, response_msg

field_name = str
default_value = Any


def get_json_values_from_req(*field_names: Union[field_name, tuple[field_name, default_value]]):
    data, is_json = g.json
    if not is_json:
        abort(response_msg("body is not json", 415))

    values, values_error = get_json_values(data, *field_names)

    if values_error:
        abort(response_msg(values_error, 400))

    return values



from .response_msg import response_msg
from .get_json_values import get_json_values

from .create_file_response import create_file_response
from .create_folder_for_file import create_folder_for_file
from .get_all_vars import get_all_values, get_all_fields
from .get_datetime_now import get_datetime_now
from .get_db_path import get_db_path
from .get_json import get_json
from .get_json_list_from_req import get_json_list_from_req
from .get_json_values_from_req import get_json_values_from_req
from .get_secret_key import get_secret_key, get_secret_key_rnd
from .import_all_tables import import_all_tables
from .ip_to_emoji import ip_to_emoji, emoji_to_ip
from .jsonify_list import jsonify_list
from .parse_date import parse_date
from .permission_required import create_permission_required_decorator
from .permission_required import permission_required, permission_required_any
from .randstr import randstr
from .register_blueprints import register_blueprints
from .response_not_found import response_not_found
from .use_db_session import use_db_session
from .use_userId import use_userId, use_userId_optional
from .use_user import use_user
from .use_user_optional import use_user_optional



def get_all_field_names(obj: object):
    return (attr for attr in dir(obj) if not callable(getattr(obj, attr)) and not attr.startswith("__"))


def get_all_values(obj: object):
    return (getattr(obj, field) for field in get_all_field_names(obj))


def get_all_fields(obj: object):
    return ((field, getattr(obj, field)) for field in get_all_field_names(obj))



def remove_user_role(userId, roleId, dev):
    userId, roleId = int(userId), int(roleId)
    print(f"remove_user_role {userId=} {roleId=} {dev=}")
    from bafser import db_session, Role
    from bafser.data.user import get_user_table

    db_session.global_init(dev)
    db_sess = db_session.create_session()
    User = get_user_table()
    user_admin = User.get_admin(db_sess)
    user = db_sess.get(User, userId)
    if not user:
        print(f"User with id [{userId}] does not exist")
        return
    role = db_sess.get(Role, roleId)
    if not role:
        print(f"Role with id [{roleId}] does not exist")
        return

    ok = user.remove_role(user_admin, roleId)

    if not ok:
        print(f"User [{user.login}] does not has [{role.name}] role")
        return

    print(f"Role [{role.name}] removed from User [{user.login}]")


def run(args):
    if not (len(args) == 2 or (len(args) == 3 and args[-1] == "dev")):
        print("remove_user_role: userId roleId [dev]")
    else:
        remove_user_role(args[0], args[1], args[-1] == "dev")



from datetime import datetime
from typing import Union

from sqlalchemy import Column, ForeignKey, Integer, orm
from sqlalchemy.orm import Session

from .. import SqlAlchemyBase
from ._tables import TablesBase


class UserRole(SqlAlchemyBase):
    __tablename__ = TablesBase.UserRole

    userId = Column(Integer, ForeignKey("User.id"), primary_key=True)
    roleId = Column(Integer, ForeignKey("Role.id"), primary_key=True)

    role = orm.relationship("Role")

    def __repr__(self):
        return f"<UserRole> user: {self.userId} role: {self.roleId}"

    def get_dict(self):
        return self.to_dict(only=("userId", "roleId"))

    @staticmethod
    def new(creator, userId: int, roleId: int, now: datetime = None, commit=True, db_sess: Session = None):
        from .. import Log
        db_sess = db_sess if db_sess else Session.object_session(creator)

        user_role = UserRole(userId=userId, roleId=roleId)
        db_sess.add(user_role)

        Log.added(user_role, creator, [
            ("userId", user_role.userId),
            ("roleId", user_role.roleId),
        ], now, commit, db_sess)
        return user_role

    @staticmethod
    def get(db_sess: Session, userId: int, roleId: int) -> Union["UserRole", None]:
        return db_sess.query(UserRole).filter(UserRole.userId == userId, UserRole.roleId == roleId).first()

    def delete(self, actor):
        from .. import Log
        db_sess = Session.object_session(self)
        db_sess.delete(self)
        Log.deleted(self, actor, [
            ("userId", self.userId),
            ("roleId", self.roleId),
        ])



import importlib
import os
import bafser_config


def register_blueprints(app):
    if not os.path.exists(bafser_config.blueprints_folder):
        return

    blueprints_module = bafser_config.blueprints_folder.replace("/", ".").replace("\\", ".")
    for file in os.listdir(bafser_config.blueprints_folder):
        if not file.endswith(".py"):
            continue
        module = blueprints_module + "." + file[:-3]
        blueprint = importlib.import_module(module).blueprint
        app.register_blueprint(blueprint)



from sqlalchemy import Boolean, Column, DefaultClause
from sqlalchemy.orm import Session

from .. import SqlAlchemyBase, SingletonMixin


class DBState(SqlAlchemyBase, SingletonMixin):
    __tablename__ = "db_state"

    initialized = Column(Boolean, DefaultClause("0"), nullable=False)

    @staticmethod
    def is_initialized(db_sess: Session):
        return DBState.get(db_sess).initialized

    @staticmethod
    def mark_as_initialized(db_sess: Session):
        DBState.get(db_sess).initialized = True
        db_sess.commit()



import os
import sys

current = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(current, "src"))

from bafser import AppConfig, create_app


app, run = create_app(__name__, AppConfig(
    FRONTEND_FOLDER="test/build",
    # PAGE404="404.html",
    MESSAGE_TO_FRONTEND="",
    DEV_MODE="dev" in sys.argv,
    DELAY_MODE="delay" in sys.argv,
))


# run(False, init_db, init_values, port=5000, host="127.0.0.1")
# run(False)
run(__name__ == "__main__", port=5000)



import importlib
import sys


def cli():
    scripts = [
        ("add_user_role", "userId roleId [dev]"),
        ("add_user", "login password name roleId [dev]"),
        ("change_user_password", "login new_password [dev]"),
        ("remove_user_role", "userId roleId [dev]"),
        ("alembic", "<init | revision>"),
    ]

    if len(sys.argv) < 2 or sys.argv[1] not in map(lambda v: v[0], scripts):
        ml = max(map(lambda v: len(v[0]), scripts))
        ml2 = max(map(lambda v: len(v[1]), scripts))
        l = ml + ml2 + 5
        t = " Scripts "
        l2 = (l - len(t))
        print("-" * (l2 // 2) + t + "-" * (l2 // 2 + l2 % 2))
        print("\n".join(map(lambda v: f"{' ' * (ml - len(v[0]))} {v[0]} : {v[1]}", scripts)))
        print("-" * l)
        return

    importlib.import_module(".scripts." + sys.argv[1], "bafser").run(sys.argv[2:])


if __name__ == "__main__":
    cli()



from functools import wraps
from bafser import db_session


def use_db_session():
    def wrapper(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            db_sess = db_session.create_session()
            try:
                return fn(*args, **kwargs, db_sess=db_sess)
            finally:
                db_sess.close()

        return decorator

    return wrapper



from sqlalchemy import Column, ForeignKey, Integer, String, orm

from .. import SqlAlchemyBase


class Permission(SqlAlchemyBase):
    __tablename__ = "Permission"

    roleId = Column(Integer, ForeignKey("Role.id"), primary_key=True)
    operationId = Column(String(32), ForeignKey("Operation.id"), primary_key=True)

    operation = orm.relationship("Operation")

    def __repr__(self):
        return f"<Permission> role: {self.roleId} oper: {self.operationId}"

    def get_dict(self):
        return self.to_dict(only=("roleId", "operationId"))



from typing import Any, Type, TypeVar

from sqlalchemy import Column, String
from sqlalchemy.orm import Session
from werkzeug.security import generate_password_hash, check_password_hash

from .. import SqlAlchemyBase, ObjMixin, UserRole
from ..utils import get_datetime_now
from ._roles import RolesBase
from ._tables import TablesBase
from .permission import Permission

T = TypeVar("T", bound="UserBase")
User: "Type[UserBase]" = None
TFieldName = str
TValue = Any


def get_user_table():
    if User is None:
        raise Exception("[bafser] No class inherited from UserBase")
    return User


class UserBase(SqlAlchemyBase, ObjMixin):
    __tablename__ = TablesBase.User

    login = Column(String(64), index=True, unique=True, nullable=False)
    password = Column(String(128), nullable=False)
    name = Column(String(64), nullable=False)

    def __repr__(self):
        return f"<{self.__class__.__name__}> [{self.id}] {self.login}"

    def __init_subclass__(cls, **kwargs):
        global User
        User = cls

    @classmethod
    def new(cls, creator: "UserBase", login: str, password: str, name: str, roles: list[int], db_sess: Session = None, **kwargs):
        from .. import Log
        db_sess = db_sess if db_sess else Session.object_session(creator)
        user, add_changes = cls._new(db_sess, {"login": login, "name": name}, **kwargs)
        user.set_password(password)
        db_sess.add(user)

        now = get_datetime_now()
        Log.added(user, creator, [
            ("login", user.login),
            ("name", user.name),
            ("password", "***"),
            *add_changes,
        ], now, db_sess=db_sess)

        for roleId in roles:
            UserRole.new(creator, user.id, roleId, now=now, commit=False, db_sess=db_sess)

        db_sess.commit()

        return user

    @classmethod
    def _new(cls: Type[T], db_sess: Session, user_kwargs: dict, **kwargs) -> tuple[T, list[tuple[TFieldName, TValue]]]:
        user = cls(**user_kwargs)
        return user, []

    @classmethod
    def get_by_login(cls, db_sess: Session, login: str, includeDeleted=False):
        return cls.query(db_sess, includeDeleted).filter(cls.login == login).first()

    @classmethod
    def create_admin(cls, db_sess: Session):
        fake_creator = UserBase.get_fake_system()
        return cls.new(fake_creator, "admin", "admin", "Admin", [RolesBase.admin], db_sess=db_sess)

    @staticmethod
    def _create_admin(db_sess: Session):
        User = get_user_table()
        admin = User.get_admin(db_sess)
        if admin:
            return admin
        return User.create_admin(db_sess)

    @classmethod
    def get_admin(cls, db_sess: Session):
        return db_sess.query(cls).join(UserRole).filter(UserRole.roleId == RolesBase.admin).first()

    _is_admin = None

    def is_admin(self):
        if self._is_admin is None:
            self._is_admin = self.has_role(RolesBase.admin)
        return self._is_admin

    @staticmethod
    def get_fake_system():
        return UserBase(id=1, name="System")

    @classmethod
    def all_of_role(cls, db_sess: Session, role: int, includeDeleted=False):
        return cls.query(db_sess, includeDeleted).join(UserRole).filter(UserRole.roleId == role).all()

    def update_password(self, actor: "UserBase", password: str):
        from .. import Log
        self.set_password(password)
        Log.updated(self, actor, [("password", "***", "***")])

    def update_name(self, actor: "UserBase", name: str):
        from .. import Log
        oldname = self.name
        self.name = name
        Log.updated(self, actor, [("name", oldname, name)])

    def set_password(self, password: str):
        self.password = generate_password_hash(password)

    def check_password(self, password: str):
        return check_password_hash(self.password, password)

    def check_permission(self, operation: tuple[str, str]):
        return operation[0] in self.get_operations()

    def add_role(self, actor: "UserBase", roleId: int):
        db_sess = Session.object_session(self)
        existing = UserRole.get(db_sess, self.id, roleId)
        if existing:
            return False

        UserRole.new(actor, self.id, roleId)
        if roleId == RolesBase.admin:
            self._is_admin = True
        return True

    def remove_role(self, actor: "UserBase", roleId: int):
        db_sess = Session.object_session(self)
        user_role = UserRole.get(db_sess, self.id, roleId)
        if not user_role:
            return False

        user_role.delete(actor)
        if roleId == RolesBase.admin:
            self._is_admin = False
        return True

    def get_roles(self):
        from .. import Role
        db_sess = Session.object_session(self)
        roles = db_sess\
            .query(Role)\
            .join(UserRole, UserRole.roleId == Role.id)\
            .filter(UserRole.userId == self.id)\
            .values(Role.name)

        return list(map(lambda v: v[0], roles))

    def has_role(self, roleId: int):
        db_sess = Session.object_session(self)
        ur = db_sess\
            .query(UserRole.roleId)\
            .filter(UserRole.roleId == roleId)\
            .filter(UserRole.userId == self.id)\
            .first()

        return ur is not None

    def get_operations(self):
        from .. import Role
        db_sess = Session.object_session(self)
        operations = db_sess\
            .query(Permission)\
            .join(Role, Permission.roleId == Role.id)\
            .join(UserRole, UserRole.roleId == Role.id)\
            .filter(UserRole.userId == self.id)\
            .values(Permission.operationId)

        return list(map(lambda v: v[0], operations))

    def get_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "login": self.login,
            "roles": self.get_roles(),
            "operations": self.get_operations(),
        }

    def get_dict_full(self):
        return {
            "id": self.id,
            "deleted": self.deleted,
            "name": self.name,
            "login": self.login,
            "roles": self.get_roles(),
            "operations": self.get_operations(),
        }



from .add_user_role import add_user_role
from .add_user import add_user
from .alembic import init as alembic_init, revision as alembic_revision
from .change_user_password import change_user_password
from .remove_user_role import remove_user_role



def change_user_password(login, password, dev):
    print(f"change_user_password {login=} {password=} {dev=}")
    from bafser import db_session
    from bafser.data.user import get_user_table

    db_session.global_init(dev)
    db_sess = db_session.create_session()
    User = get_user_table()
    user = User.get_by_login(db_sess, login, includeDeleted=True)
    if user is None:
        print("User does not exist")
        return
    user.set_password(password)
    db_sess.commit()
    print("Password changed")


def run(args):
    if not (len(args) == 2 or (len(args) == 3 and args[-1] == "dev")):
        print("change_user_password: login new_password [dev]")
    else:
        change_user_password(args[0], args[1], args[-1] == "dev")



import importlib
import os
import bafser_config


def import_all_tables():
    from ..data import image
    from ..data import log
    from ..data import operation
    from ..data import permission
    from ..data import role
    from ..data import user_role

    if not os.path.exists(bafser_config.data_tables_folder):
        return

    data_tables_module = bafser_config.data_tables_folder.replace("/", ".").replace("\\", ".")
    for file in os.listdir(bafser_config.data_tables_folder):
        if not file.endswith(".py"):
            continue
        module = data_tables_module + "." + file[:-3]
        importlib.import_module(module)



from datetime import datetime, timedelta
import json
import logging
import logging.handlers
import os

from flask import g, has_request_context, request

from bafser import create_folder_for_file, ip_to_emoji
import bafser_config


def customTime(*args):
    utc_dt = datetime.utcnow()
    utc_dt += timedelta(hours=3)
    return utc_dt.timetuple()


class InfoFilter(logging.Filter):
    def filter(self, rec):
        return rec.levelno == logging.INFO and rec.name == "root"


class RequestFormatter(logging.Formatter):
    converter = customTime
    max_msg_len = -1
    max_json_len = 2048
    json_indent = None
    outer_args: list[str] = []

    def format(self, record):
        if has_request_context():
            url_start = request.url.find(bafser_config.api_url)
            record.url = request.url[url_start:] if url_start >= 0 else request.url
            record.method = request.method
            remote_addr = request.headers.get("X-Real-IP", request.remote_addr)
            record.ip = remote_addr
            record.ip_emoji = ip_to_emoji(remote_addr)
            record.req_id = g.get("req_id", "")
            record.uid = g.get("userId", "")
            g_json = g.get("json", None)
            if g_json is not None and g_json[1]:
                record.json = json.dumps(g_json[0], indent=self.json_indent)
            else:
                record.json = "[no json]"
        else:
            record.url = "[url]"
            record.method = "[method]"
            record.ip = "[ip]"
            record.ip_emoji = "[ip_emoji]"
            record.req_id = "[req_id]"
            record.json = "[json]"
            record.uid = "[uid]"

        if self.max_msg_len > 0 and len(record.msg) > self.max_msg_len:
            record.msg = record.msg[:self.max_msg_len] + "..."

        if self.max_json_len > 0 and len(record.json) > self.max_json_len:
            record.json = record.json[:1024] + "..."

        for arg in self.outer_args:
            setattr(record, arg, record.args.get(arg, f"[{arg}]"))

        return super().format(record)


class RotatingFileHandler(logging.handlers.RotatingFileHandler):
    def __init__(self, *kargs, **kwargs):
        self.rollover = False
        super().__init__(*kargs, **kwargs)

    def doRollover(self):
        self.rollover = True
        s = self.stream
        self.stream = self._open()
        if s:
            s.close()

    def _open(self):
        filename = get_log_fpath(self.baseFilename, self.rollover)
        self.rollover = False
        bn = self.baseFilename
        self.baseFilename = filename
        r = super()._open()
        self.baseFilename = bn
        return r


def get_log_fpath(fpath: str, next=False):
    i = 0
    n = fpath.split(".")
    name, ext = (".".join(n[:-1]), n[-1]) if len(n) > 1 else (n[0], "")
    while True:
        i += 1
        npath = f"{name}.{i}.{ext}"
        if not os.path.exists(npath):
            if next:
                return npath
            if i - 1 == 0:
                return fpath
            return f"{name}.{i - 1}.{ext}"


MaxBytes = 8 * 1000 * 1000


def setLogging():
    logging.basicConfig(
        level=logging.DEBUG,
        # filename="log.log",
        format="[%(asctime)s] %(levelname)s in %(module)s (%(name)s): %(message)s",
        encoding="utf-8"
    )
    create_folder_for_file(bafser_config.log_errors_path)
    create_folder_for_file(bafser_config.log_info_path)
    create_folder_for_file(bafser_config.log_requests_path)
    create_folder_for_file(bafser_config.log_frontend_path)
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    logger.handlers.clear()
    logging.Formatter.converter = customTime

    formatter_error = RequestFormatter("[%(asctime)s] %(ip_emoji)s (%(req_id)s by uid=%(uid)-6s) %(method)-6s %(url)-40s | %(levelname)s in %(module)s (%(name)s):\nReq json: %(json)s\n%(message)s\n")  # noqa: E501
    formatter_error.max_json_len = -1
    file_handler_error = RotatingFileHandler(
        bafser_config.log_errors_path, mode="a", encoding="utf-8", maxBytes=MaxBytes)
    file_handler_error.setFormatter(formatter_error)
    file_handler_error.setLevel(logging.WARNING)
    file_handler_error.encoding = "utf-8"
    logger.addHandler(file_handler_error)

    formatter_info = RequestFormatter("%(req_id)s;%(ip_emoji)s;%(uid)-6s;%(asctime)s;%(method)-4s;%(url)s;%(module)s;%(message)s")
    formatter_info.max_json_len = 4096
    file_handler_info = RotatingFileHandler(
        bafser_config.log_info_path, mode="a", encoding="utf-8", maxBytes=MaxBytes)
    file_handler_info.setFormatter(formatter_info)
    file_handler_info.addFilter(InfoFilter())
    file_handler_info.encoding = "utf-8"
    logger.addHandler(file_handler_info)

    logger_requests = get_logger_requests()
    logger_requests.setLevel(logging.DEBUG)
    formatter_req = RequestFormatter("%(req_id)s;%(ip_emoji)s;%(uid)-6s;%(asctime)s;%(method)-4s;%(url)s;%(message)s")
    formatter_req.max_msg_len = 1024
    file_handler_req = RotatingFileHandler(
        bafser_config.log_requests_path, mode="a", encoding="utf-8", maxBytes=MaxBytes)
    file_handler_req.setFormatter(formatter_req)
    file_handler_req.setLevel(logging.INFO)
    file_handler_req.encoding = "utf-8"
    logger_requests.addHandler(file_handler_req)

    logger_frontend = get_logger_frontend()
    logger_frontend.setLevel(logging.DEBUG)
    formatter_frontend = RequestFormatter("[%(asctime)s] %(ip_emoji)s (uid=%(uid)s):\n%(json)s\n%(message)s\n")
    formatter_frontend.max_json_len = 8192
    formatter_frontend.json_indent = 4
    file_handler_frontend = RotatingFileHandler(
        bafser_config.log_frontend_path, mode="a", encoding="utf-8", maxBytes=MaxBytes)
    file_handler_frontend.setFormatter(formatter_frontend)
    file_handler_frontend.setLevel(logging.INFO)
    file_handler_frontend.encoding = "utf-8"
    logger_frontend.addHandler(file_handler_frontend)


def get_logger_frontend():
    return logging.getLogger("frontend")


def get_logger_requests():
    return logging.getLogger("requests")


def log_frontend_error():
    get_logger_frontend().info("")


def add_file_logger(
    fpath: str,
    name: str,
    format="%(req_id)s;%(ip_emoji)s;%(uid)-6s;%(asctime)s;%(method)s;%(url)s;%(levelname)s;%(module)s;%(message)s",
    outer_args: list[str] = [],
    max_json_len=4096
):
    create_folder_for_file(fpath)
    logger = logging.getLogger(name)
    formatter = RequestFormatter(format)
    formatter.max_json_len = max_json_len
    formatter.outer_args = outer_args
    file_handler = RotatingFileHandler(fpath, mode="a", encoding="utf-8", maxBytes=MaxBytes)
    file_handler.setFormatter(formatter)
    file_handler.encoding = "utf-8"
    logger.addHandler(file_handler)
    return logger


class ParametrizedLogger:
    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def _get_args(self):
        return {}

    def info(self, msg: str):
        self._log(self.logger.info, msg)

    def error(self, msg: str):
        self._log(self.logger.error, msg)

    def warning(self, msg: str):
        self._log(self.logger.warning, msg)

    def _log(self, fn, msg: str):
        fn(msg, self._get_args(), stacklevel=3)



"""v1

Revision ID: 1051ea08f476
Revises: 
Create Date: 2025-08-11 20:42:07.214630

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '1051ea08f476'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('Log',
    sa.Column('date', sa.DateTime(), nullable=False),
    sa.Column('actionCode', sa.String(length=16), nullable=False),
    sa.Column('userId', sa.Integer(), nullable=False),
    sa.Column('userName', sa.String(length=64), nullable=False),
    sa.Column('tableName', sa.String(length=16), nullable=False),
    sa.Column('recordId', sa.Integer(), nullable=False),
    sa.Column('changes', sa.JSON(), nullable=False),
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_Log')),
    sa.UniqueConstraint('id', name=op.f('uq_Log_id')),
    mysql_charset='utf8mb4',
    mysql_collate='utf8mb4_unicode_ci'
    )
    op.create_table('Operation',
    sa.Column('id', sa.String(length=32), nullable=False),
    sa.Column('name', sa.String(length=32), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_Operation')),
    sa.UniqueConstraint('id', name=op.f('uq_Operation_id')),
    mysql_charset='utf8mb4',
    mysql_collate='utf8mb4_unicode_ci'
    )
    op.create_table('Role',
    sa.Column('name', sa.String(length=32), nullable=False),
    sa.Column('deleted', sa.Boolean(), server_default='0', nullable=False),
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_Role')),
    sa.UniqueConstraint('id', name=op.f('uq_Role_id')),
    mysql_charset='utf8mb4',
    mysql_collate='utf8mb4_unicode_ci'
    )
    op.create_table('User',
    sa.Column('login', sa.String(length=64), nullable=False),
    sa.Column('password', sa.String(length=128), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=False),
    sa.Column('deleted', sa.Boolean(), server_default='0', nullable=False),
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_User')),
    sa.UniqueConstraint('id', name=op.f('uq_User_id')),
    mysql_charset='utf8mb4',
    mysql_collate='utf8mb4_unicode_ci'
    )
    with op.batch_alter_table('User', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_User_login'), ['login'], unique=True)

    op.create_table('Image',
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.Column('type', sa.String(length=16), nullable=False),
    sa.Column('creationDate', sa.DateTime(), nullable=False),
    sa.Column('deletionDate', sa.DateTime(), nullable=True),
    sa.Column('createdById', sa.Integer(), nullable=False),
    sa.Column('deleted', sa.Boolean(), server_default='0', nullable=False),
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.ForeignKeyConstraint(['createdById'], ['User.id'], name=op.f('fk_Image_createdById_User')),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_Image')),
    sa.UniqueConstraint('id', name=op.f('uq_Image_id')),
    mysql_charset='utf8mb4',
    mysql_collate='utf8mb4_unicode_ci'
    )
    op.create_table('Permission',
    sa.Column('roleId', sa.Integer(), nullable=False),
    sa.Column('operationId', sa.String(length=32), nullable=False),
    sa.ForeignKeyConstraint(['operationId'], ['Operation.id'], name=op.f('fk_Permission_operationId_Operation')),
    sa.ForeignKeyConstraint(['roleId'], ['Role.id'], name=op.f('fk_Permission_roleId_Role')),
    sa.PrimaryKeyConstraint('roleId', 'operationId', name=op.f('pk_Permission')),
    mysql_charset='utf8mb4',
    mysql_collate='utf8mb4_unicode_ci'
    )
    op.create_table('UserRole',
    sa.Column('userId', sa.Integer(), nullable=False),
    sa.Column('roleId', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['roleId'], ['Role.id'], name=op.f('fk_UserRole_roleId_Role')),
    sa.ForeignKeyConstraint(['userId'], ['User.id'], name=op.f('fk_UserRole_userId_User')),
    sa.PrimaryKeyConstraint('userId', 'roleId', name=op.f('pk_UserRole')),
    mysql_charset='utf8mb4',
    mysql_collate='utf8mb4_unicode_ci'
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('UserRole')
    op.drop_table('Permission')
    op.drop_table('Image')
    with op.batch_alter_table('User', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_User_login'))

    op.drop_table('User')
    op.drop_table('Role')
    op.drop_table('Operation')
    op.drop_table('Log')
    # ### end Alembic commands ###



from typing import Any, Union

field_name = str
default_value = Any
values = Union[list[Any], Any]
error = str


def get_json_values(d: dict, *field_names: Union[field_name, tuple[field_name, default_value]]) -> tuple[values, error]:
    r = []
    for field in field_names:
        if isinstance(field, tuple) and len(field) == 2:
            field_name, default_value = field
            have_default = True
        else:
            field_name = field
            have_default = False

        if field_name in d:
            r.append(d[field_name])
        elif have_default:
            r.append(default_value)
        else:
            r = None if len(field_names) == 1 else list(map(lambda _: None, field_names))
            return r, f"{field_name} is undefined"
    if len(r) == 1:
        return r[0], None
    return r, None



from ..utils import get_all_fields


class TablesBase:
    User = "User"
    Role = "Role"
    UserRole = "UserRole"
    Image = "Image"

    @classmethod
    def get_all(cls):
        return list(get_all_fields(cls()))



# base for flask server by Mixel Te

from .utils.response_msg import response_msg
from .utils.get_json_values import get_json_values

from .utils.create_file_response import create_file_response
from .utils.create_folder_for_file import create_folder_for_file
from .utils.get_datetime_now import get_datetime_now
from .utils.get_json import get_json
from .utils.get_json_list_from_req import get_json_list_from_req
from .utils.get_json_values_from_req import get_json_values_from_req
from .utils.ip_to_emoji import ip_to_emoji, emoji_to_ip
from .utils.jsonify_list import jsonify_list
from .utils.parse_date import parse_date
from .utils.permission_required import create_permission_required_decorator
from .utils.permission_required import permission_required, permission_required_any
from .utils.randstr import randstr
from .utils.response_not_found import response_not_found
from .utils.use_db_session import use_db_session
from .utils.use_userId import use_userId, use_userId_optional
from .utils.use_user import use_user
from .utils.use_user_optional import use_user_optional

from .app import AppConfig, create_app
from .logger import get_logger_frontend, log_frontend_error, get_log_fpath, add_file_logger, ParametrizedLogger

from .db_session import SqlAlchemyBase
from .table_base import TableBase, IdMixin, ObjMixin, SingletonMixin
from .data._tables import TablesBase
from .data._roles import RolesBase
from .data.operation import OperationsBase
from .data.user_role import UserRole
from .data.user import UserBase
from .data.log import Log
from .data.role import Role
from .data.image import Image


class M:
    DELETE = "DELETE"
    GET = "GET"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    PATCH = "PATCH"
    POST = "POST"
    PUT = "PUT"



from functools import wraps
from flask import abort
from flask_jwt_extended import get_jwt_identity, unset_jwt_cookies
from sqlalchemy.orm import Session

from . import response_msg


def use_user():
    from ..data.user import get_user_table

    def wrapper(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            if "db_sess" not in kwargs:
                abort(500, "use_user: no db_sess")
            db_sess: Session = kwargs["db_sess"]

            jwt_identity = get_jwt_identity()
            if (not isinstance(jwt_identity, list) and not isinstance(jwt_identity, tuple)) or len(jwt_identity) != 2:
                response = response_msg("The JWT has expired")
                unset_jwt_cookies(response)
                return response, 401

            user = get_user_table().get(db_sess, jwt_identity[0])
            if not user:
                return response_msg("User not found"), 401
            if user.password != jwt_identity[1]:
                response = response_msg("The JWT has expired")
                unset_jwt_cookies(response)
                return response, 401

            return fn(*args, **kwargs, user=user)

        return decorator

    return wrapper



from bafser import RolesBase
from test.data._operations import Operations


class Roles(RolesBase):
    user = 2


Roles.ROLES = {
    Roles.user: {
        "name": "User",
        "operations": []
    },
}



import os
import uuid


def get_secret_key_rnd(path):
    if os.path.exists(path):
        with open(path, "r", encoding="utf8") as f:
            return f.read()
    else:
        with open(path, "w", encoding="utf8") as f:
            key = str(uuid.uuid4())
            f.write(key)
            return key


def get_secret_key(path):
    if os.path.exists(path):
        with open(path, "r", encoding="utf8") as f:
            return f.read()
    else:
        raise Exception(f"Secret key not found! [{path}]")



from functools import wraps
from flask_jwt_extended import get_jwt_identity, unset_jwt_cookies

from . import response_msg


def use_userId():
    def wrapper(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            jwt_identity = get_jwt_identity()
            if (not isinstance(jwt_identity, list) and not isinstance(jwt_identity, tuple)) or len(jwt_identity) != 2:
                response = response_msg("The JWT has expired")
                unset_jwt_cookies(response)
                return response, 401

            userId = jwt_identity[0]
            return fn(*args, **kwargs, userId=userId)

        return decorator

    return wrapper


def use_userId_optional():
    def wrapper(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            try:
                jwt_identity = get_jwt_identity()
                if (not isinstance(jwt_identity, list) and not isinstance(jwt_identity, tuple)) or len(jwt_identity) != 2:
                    response = response_msg("The JWT has expired")
                    unset_jwt_cookies(response)
                    return response, 401
                userId = jwt_identity[0]
            except Exception:
                userId = None

            return fn(*args, **kwargs, userId=userId)

        return decorator

    return wrapper



import os

from alembic.config import Config
from alembic import command

from .utils import get_db_path
import bafser_config


def create_alembic_config(dev: bool):
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", bafser_config.migrations_folder)
    alembic_cfg.set_main_option("file_template", "%%(year)d_%%(month).2d_%%(day).2d_%%(rev)s_%%(slug)s")

    if dev:
        db_path = bafser_config.db_dev_path
        issqlite = True
    else:
        db_path = get_db_path(bafser_config.db_path)
        issqlite = not bafser_config.db_mysql
    alembic_cfg.set_main_option("issqlite", "1" if issqlite else "0")

    if issqlite:
        alembic_cfg.set_main_option("sqlalchemy.url", f"sqlite:///{db_path}?check_same_thread=False")
        if not os.path.exists(db_path):
            dirname = os.path.dirname(db_path)
            if dirname != "":
                os.makedirs(dirname, exist_ok=True)
            with open(db_path, "w"):
                pass
    else:
        alembic_cfg.set_main_option("sqlalchemy.url", f"mysql+pymysql://{db_path}?charset=UTF8mb4")
    return alembic_cfg


def alembic_upgrade(dev: bool):
    alembic_cfg = create_alembic_config(dev)
    command.upgrade(alembic_cfg, "head")


def run():
    from logging.config import fileConfig
    from sqlalchemy import engine_from_config
    from sqlalchemy import pool

    from alembic import context
    config = context.config
    if config.config_file_name is not None:
        fileConfig(config.config_file_name)

    from .db_session import SqlAlchemyBase
    from .utils.import_all_tables import import_all_tables
    import_all_tables()
    target_metadata = SqlAlchemyBase.metadata

    issqlite = config.get_main_option("issqlite") == "1"

    def run_migrations_offline() -> None:
        """Run migrations in 'offline' mode.

        This configures the context with just a URL
        and not an Engine, though an Engine is acceptable
        here as well.  By skipping the Engine creation
        we don't even need a DBAPI to be available.

        Calls to context.execute() here emit the given string to the
        script output.

        """
        url = config.get_main_option("sqlalchemy.url")
        context.configure(
            url=url,
            target_metadata=target_metadata,
            literal_binds=True,
            dialect_opts={"paramstyle": "named"},
            render_as_batch=issqlite,
        )

        with context.begin_transaction():
            context.run_migrations()

    def run_migrations_online() -> None:
        """Run migrations in 'online' mode.

        In this scenario we need to create an Engine
        and associate a connection with the context.

        """
        connectable = engine_from_config(
            config.get_section(config.config_ini_section, {}),
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )

        with connectable.connect() as connection:
            context.configure(
                connection=connection,
                target_metadata=target_metadata,
                render_as_batch=issqlite,
            )

            with context.begin_transaction():
                context.run_migrations()

    if context.is_offline_mode():
        run_migrations_offline()
    else:
        run_migrations_online()



from sqlalchemy import Column, orm, String
from sqlalchemy.orm import Session

from .. import SqlAlchemyBase, ObjMixin, get_datetime_now
from ._roles import RoleDesc, RolesBase, TRole, get_roles
from ._tables import TablesBase
from .log import Actions, Log
from .operation import Operation, get_operations
from .permission import Permission


class Role(SqlAlchemyBase, ObjMixin):
    __tablename__ = TablesBase.Role

    name = Column(String(32), nullable=False)

    permissions = orm.relationship("Permission")

    def __repr__(self):
        return f"<Role> [{self.id}] {self.name}"

    def get_dict(self):
        return self.to_dict(only=("id", "name"))

    @staticmethod
    def update_roles_permissions(db_sess: Session):
        ROLES = get_roles().ROLES
        # add new operations
        operations_new = get_operations().get_all()
        operations_new_ids = [v[0] for v in operations_new]
        operations_cur = list(db_sess.query(Operation).all())
        operations_cur_ids = [v.id for v in operations_cur]
        for operation in operations_new:
            operation_id, operation_name = operation
            if operation_id not in operations_cur_ids:
                db_sess.add(Operation(id=operation_id, name=operation_name))

        # update roles
        roles_new_ids = ROLES.keys()
        roles = db_sess.query(Role).all()
        roles_cur_ids = [v.id for v in roles]
        removed_roles: list[int] = []
        restored_roles: list[int] = []
        updated_roles: list[tuple[int, str, str]] = []
        for role in list(roles):
            # delete removed roles
            if role.id not in roles_new_ids and role.id != RolesBase.admin:
                role.deleted = True
                removed_roles.append(role.id)
                for permission in list(role.permissions):
                    db_sess.delete(permission)
                continue

            # restore role with old id
            if role.deleted:
                role.deleted = False
                restored_roles.append(role.id)

            # get role data
            if role.id == RolesBase.admin:
                name = "Admin"
                operations = operations_new
            else:
                name = ROLES[role.id]["name"]
                operations = ROLES[role.id]["operations"]
            operations_ids = [v[0] for v in operations]

            # update role name
            if role.name != name:
                updated_roles.append((role.id, role.name, name))
                role.name = name

            # delete removed permissions
            cur_permissions: list[Permission] = list(role.permissions)
            for permission in cur_permissions:
                if permission.operationId not in operations_ids:
                    db_sess.delete(permission)

            # add new permissions
            cur_operations_ids = [v.operationId for v in cur_permissions]
            for operation in operations:
                operation_id = operation[0]
                if operation_id not in cur_operations_ids:
                    db_sess.add(Permission(roleId=role.id, operationId=operation_id))

        # delete removed operations
        for operation in operations_cur:
            if operation.id not in operations_new_ids:
                db_sess.delete(operation)

        # add new roles
        new_roles: list[tuple[int, str]] = []
        admin_role: tuple[TRole, RoleDesc] = (RolesBase.admin, {"name": "Admin", "operations": operations_new})
        for role_id, role_data in [admin_role, *list(ROLES.items())]:
            if role_id in roles_cur_ids:
                continue
            name, operations = role_data["name"], role_data["operations"]

            role = Role(name=name, id=role_id)
            new_roles.append((role_id, name))
            db_sess.add(role)

            for operation in operations:
                db_sess.add(Permission(roleId=role_id, operationId=operation[0]))

        now = get_datetime_now()

        def log_role(actionCode, recordId, changes):
            db_sess.add(Log(
                date=now,
                actionCode=actionCode,
                userId=1,
                userName="System",
                tableName=TablesBase.Role,
                recordId=recordId,
                changes=changes
            ))

        for role_id in removed_roles:
            log_role(Actions.deleted, role_id, [])
        for role_id in restored_roles:
            log_role(Actions.restored, role_id, [])
        for (role_id, old_name, name) in updated_roles:
            log_role(Actions.updated, role_id, [("name", old_name, name)])
        for (role_id, name) in new_roles:
            log_role(Actions.added, role_id, [("name", None, name)])

        db_sess.commit()
